#if !(BufferedCAN || QueuedCAN)

/*
 * CAN implementation for the primary (master) CAN controller.
 */

#include "config.h"
#include "stm32/stm32-can.h"

#ifndef InOrderCAN
#define InOrderCAN	1
#endif

//-----------------------------------------------------------------------------

#if 1
  #define trace(x)	printf x
  #define trace_on
#else
  #define trace(x)
#endif

#if 0
  #define trace2(x)	printf x
  #define trace2_on
#else
  #define trace2(x)
#endif

//-----------------------------------------------------------------------------

#ifndef STM32F10X_CL
  #define CAN_BANKS	14
#else
  #define CAN_BANKS	28
#endif

// Filter modes:
#define Mask16		0
#define List16		1
#define Mask32		2
#define List32		3
#define Init		-1

static int8_t	fmode;		// last filter mode
static uint8_t	fbank;		// last bank
static uint8_t	fslot;		// last slot within bank
static uint8_t	fqueue;		// last queue number
static uint8_t	fnumber[2];	// last per-queue filter number (base)

//-----------------------------------------------------------------------------

void
canInit(enum CanMode mode)
{
    trace(("canInit %d\n", mode));
    BBPeriphMask(RCC->APB1ENR, RCC_APB1ENR_CAN1EN) = 1;	// enable clock

    CAN1->MCR = CAN_MCR_RESET;				// master reset
    while (CAN1->MCR & CAN_MCR_RESET) ;

    CAN1->MCR = CAN_MCR_INRQ;				// enter initialization mode
    while (!(CAN1->MSR & CAN_MSR_INAK)) ;

    CAN1->MCR = CAN_MCR_TTCM				// time triggered mode
	      | CAN_MCR_ABOM 				// automatic bus-off management
	      | (InOrderCAN ? CAN_MCR_TXFP : 0)		// Tx priority
	      | CAN_MCR_INRQ;				// still initialization mode

    CAN1->FMR = CAN_FMR_FINIT;				// filters in initialization mode

    unsigned btr = CAN_RATEDEF;				// bit rate and mode
    if (mode == CanListen)
	btr |= CAN_BTR_SILM;
    if (mode == CanLoopback)
	btr |= CAN_BTR_LBKM;
    if (mode == CanLoopbackOff)
	btr |= CAN_BTR_SILM|CAN_BTR_LBKM;
    CAN1->BTR = btr;

    CAN1->IER = CAN_IER_TMEIE;				// enable Tx interrupt
    nvic_set_priority(USB_HP_CAN1_TX_IRQn, IrqLevel_CANtx);
    nvic_enable(USB_HP_CAN1_TX_IRQn);

    fmode = Init;
}

//-----------------------------------------------------------------------------

static int
canFilterMode(unsigned mode, unsigned qno)
{
    //			    Mask16 List16 Mask32 List32
    static char slots[] = { 2,     4,     1,     2      };

    trace(("mode=%u, ", mode));
    if (fmode != mode || fqueue != qno || fslot == slots[fmode]-1) {
	if (fbank == CAN_BANKS-1) {
	    trace(("overflow\n"));
	    return 0;
	}

	if (fmode == Init) {
	    fbank = 0;
	    fnumber[0] = 0;
	    fnumber[1] = 0;
	} else {
	    fbank += 1;
	    fnumber[fqueue] += slots[fmode];
	}
	fmode  = mode;
	fslot  = 0;
	fqueue = qno;

	BBPeriphBit(CAN1->FM1R,  fbank) = mode;
	BBPeriphBit(CAN1->FS1R,  fbank) = mode >> 1;
	BBPeriphBit(CAN1->FFA1R, fbank) = qno;
	BBPeriphBit(CAN1->FA1R,  fbank) = 1;

	trace(("advance, "));
    } else {
	fslot += 1;
    }

    trace(("bank=%u slot=%u", fbank, fslot));
    return 1;
}

static int
canFilterNumber(unsigned qno)
{
    int n = fnumber[qno] + fslot;
    trace((", number=%d\n", n));
    trace2(("\tFM1R=%08lX FS1R=%08lX FFA1R=%08lX FA1R=%08lX FR1=%08lX FR2=%08lX\n",
	   CAN1->FM1R,
	   CAN1->FS1R,
	   CAN1->FFA1R,
	   CAN1->FA1R,
	   CAN1->sFilterRegister[fbank].FR1,
	   CAN1->sFilterRegister[fbank].FR2));
    return n;
}

int
canFilterID(unsigned qno, canid_t id)
{
    trace(("canFilterID qno=%d id=%08lX\n\t", qno, id));

    int mode = IS_EID(id) ? List32 : List16;
    if (!canFilterMode(mode, qno))
	return -1;

    CAN_FilterRegister_TypeDef *p = &CAN1->sFilterRegister[fbank]; // no 16bit access
    if (mode == List16) {
	id &= SID_MASK;
	switch (fslot) {
	case 0: p->FR1 = id | id >> 16;		// fall thru
	case 1: p->FR1 = p->FR1 & 0xFFFF | id;	// fall thru
	case 2: p->FR2 = id | id >> 16;		// fall thru
	case 3: p->FR2 = p->FR2 & 0xFFFF | id;	// fall thru
	}
    } else {
	id = id & EID_MASK | CAN_IDR_IDE;
	p->FR2 = id;
	if (fslot == 0)
	    p->FR1 = id;
    }

    return canFilterNumber(qno);
}

int
canFilterMask(unsigned qno, canid_t id, canid_t mask)
{
    trace(("canFilterMask qno=%d id=%08lX mask=%08lX\n\t", qno, id, mask));

    int mode = IS_EID(id) ? Mask32 : Mask16;
    if (!canFilterMode(mode, qno))
	return -1;

    CAN_FilterRegister_TypeDef *p = &CAN1->sFilterRegister[fbank];
    if (mode == Mask16) {
	id = id & SID_MASK | (mask & SID_MASK) << 16;
	p->FR2 = id;
	if (fslot == 0)
	    p->FR1 = id;
    } else {
	p->FR1 = id   & EID_MASK | CAN_IDR_IDE;
	p->FR2 = mask & EID_MASK | CAN_IDR_IDE;
    }

    return canFilterNumber(qno);
}

//-----------------------------------------------------------------------------

void
canStart(void)
{
    trace(("canStart\n"));
    CAN1->FMR = 0;					// filters active
    BBPeriphMask(CAN1->MCR, CAN_MCR_INRQ) = 0;		// leave initialization mode
    BBPeriphMask(CAN1->MCR, CAN_MCR_SLEEP) = 0;		// don't sleep
}

void
canStop(void)
{
    trace(("canStop\n"));
    BBPeriphMask(CAN1->MCR, CAN_MCR_SLEEP) = 1;	    // sleep
}

//-----------------------------------------------------------------------------

unsigned
canRxAvail(unsigned qno)
{
    unsigned rfr = qno ? CAN1->RF1R : CAN1->RF0R;
    return CAN_RFR_FMPget(rfr);
}

unsigned
canReceive(canmsg_t *msg, unsigned qno)
{
    trace(("canReceive qno=%d\n", qno));

    qno &= 1;

    if (CAN_RFR_FMPget(CAN1->RF0R) == 0)
	return 0;
    CAN_FIFOMailBox_TypeDef *p = &CAN1->sFIFOMailBox[qno];

    unsigned idr = p->RIR;
    msg->id = idr & EID_MASK;
    if (idr & CAN_IDR_IDE)
	msg->id |= 1;
    unsigned dtr = p->RDTR;
    msg->filter = CAN_DTR_FMIget(dtr);
    msg->length = CAN_DTR_DCLget(dtr);
    msg->time = CAN_DTR_TIMEget(dtr);
    if (idr & CAN_IDR_RTR) {
	msg->flags = CANMSG_RTR;
    } else {
	msg->flags = 0;
	msg->dw[0] = p->RDLR;
	msg->dw[1] = p->RDHR;
    }
    trace2(("\tRIR=%08lX RDTR=%08lX RDLR=%08lX RDHR=%08lX\n", p->RIR, p->RDTR, p->RDLR, p->RDHR));

    if (qno) {
	CAN1->RF1R = CAN_RFR_RFOM | CAN_RFR_FOVR; // release mailbox, reset overrun status
    } else {
	CAN1->RF0R = CAN_RFR_RFOM | CAN_RFR_FOVR; // release mailbox, reset overrun status
    }

#ifdef trace_on
    trace(("\tid=%08lX filter=%u time=%u flags=%u length=%u data=",
	   msg->id, msg->filter, msg->time, msg->flags, msg->length));
    for (unsigned i = 0; i < msg->length; ++i)
	trace(("%02X ", msg->data[i]));
    trace(("\n"));
#endif
    return 1;
}

//-----------------------------------------------------------------------------

unsigned
canTxAvail(void)
{
    //			  000 001 010 011 100 101 110 111
    static char map[] = { 0,  1,  1,  2,  1,  2,  2,  3 };
    return map[CAN_TSR_TMEget(CAN1->TSR)];
}

unsigned
canTransmit(canmsg_t *msg)
{
#ifdef trace_on
    trace(("canTransmit id=%08lX flags=%u length=%u data=",
	   msg->id, msg->flags, msg->length));
    for (unsigned i = 0; i < msg->length; ++i)
	trace(("%02X ", msg->data[i]));
    trace(("\n"));
#endif

    unsigned n = CAN1->TSR;
    if (!(n & CAN_TSR_TME)) {
	trace(("\tTSR=%X\n", n));
	return 0;
    }
    n = CAN_TSR_CODEget(n);
    CAN_TxMailBox_TypeDef *p = &CAN1->sTxMailBox[n];

    unsigned idr = msg->id & EID_MASK | CAN_IDR_TXRQ;
    if (IS_EID(msg->id))
	idr |= CAN_IDR_IDE;
    if (msg->flags & CANMSG_RTR) {
	idr |= CAN_IDR_RTR;
    } else {
	p->TDHR = msg->dw[1];
	p->TDLR = msg->dw[0];
    }
    p->TDTR = msg->length;
    trace2(("\tTIR=%08lX TDTR=%08lX TDLR=%08lX TDHR=%08lX\n", idr, p->TDTR, p->TDLR, p->TDHR));
    p->TIR = idr; // start transmission

    trace(("\tmailbox %d\n", 1+n));
    return 1+n;
}

void
canAbort(unsigned box)
{
    trace(("canAbort %u\n", box));
    if (box >= 1 && box <= 3)
	CAN1->TSR = CAN_TSR_ABRQn(box-1);
    else
	CAN1->TSR = CAN_TSR_ABRQ0 | CAN_TSR_ABRQ1 | CAN_TSR_ABRQ2;
}

//-----------------------------------------------------------------------------

uint32_t
canError(void)
{
    return CAN1->ESR;
}

//-----------------------------------------------------------------------------

void __attribute__((weak))
canTxComplete(unsigned box)
{
}

void
USB_HP_CAN1_TX_IRQHandler(void)
{
    unsigned tsr = CAN1->TSR;
    if (tsr & CAN_TSR_RQCP0)
	canTxComplete(1);
    if (tsr & CAN_TSR_RQCP1)
	canTxComplete(2);
    if (tsr & CAN_TSR_RQCP2)
	canTxComplete(3);
    CAN1->TSR = tsr & (CAN_TSR_RQCP0 | CAN_TSR_RQCP1 | CAN_TSR_RQCP2);
}

// startup V1:
void USB_HP_CAN_TX_IRQHandler(void) { USB_HP_CAN1_TX_IRQHandler(); }

//-----------------------------------------------------------------------------

#endif // !(BufferedCAN || QueuedCAN)
