#ifndef __STM32_CAN_H
#define __STM32_CAN_H

#ifdef __cplusplus
 extern "C" {
#endif

//-----------------------------------------------------------------------------
#ifndef BufferedCAN

/******************************************************************************
 * Interface to STM32 integrated CAN controller.
 * Primary channel only.
 *
 * Expects:
 *	F_APB1		CAN unit clock
 *	CAN_RATEDEF	CAN_BTR(...) see below
 *	InOrderCAN	1: Tx in request order, 0: Tx in ID order
 *	IrqLevel_CANtx
 *
 * /16 mode:	Nq=16, Tseg1=11, Tseg2=4, sampled at 75%.
 * #define CAN_RATEDEF CAN_BTR(F_APB1 / (16 * CAN_RATE)), 1, 11, 4)
 *
 * /8 mode:	Nq=8, Tseg1=11, Tseg2=4, sampled at 75%.
 * #define CAN_RATEDEF CAN_BTR(F_APB1 / (8 * CAN_RATE)), 1, 5, 2)
 *
 ******************************************************************************/

typedef uint32_t canid_t;
#define CAN_SID(id)	((uint32_t)((id) & ((1<<11)-1))<<(32-11))
#define CAN_EID(id)	((uint32_t)((id) & ((1<<29)-1))<<(32-29) | 1)

#define SID_MASK	0xFFE00000
#define EID_MASK	0xFFFFFFF8

#define IS_EID(id)	((id) & 1)
#define SID_OF(id)	(((id) >> (32-11)) & ((1<<11)-1))
#define EID_OF(id)	((id) >> (32-29))

#define CANMSG_RTR	0x01	// RTR request

typedef struct canmsg_t {
    canid_t	 id;		// see above
    uint8_t	 filter : 7,	// filter id (on receive)
		 flags  : 1;	// CANMSG_...
    uint8_t	 length;	// 0..8
    uint16_t	 time;		// time code
    union {
	uint8_t	 data[8];	// byte access
	uint32_t dw[2];		// word access
    };
} canmsg_t;

enum CanMode { CanNormal, CanListen, CanLoopback, CanLoopbackOff };

// receive queue:    qno = 0..1
// transmit mailbox: box = 1..3
// filter:           bank = 0..27 for connectivity line, 0..13 otherwise

// initialize: canInit -> canFilterXXX -> canStart

void		canInit (enum CanMode mode);

     /*
      * Pack filters into subsequent slots, with padding (repetition)
      * when necessary. Return filter number - not identical to bank
      * number. Return -1 when out of banks. Do not interleave
      * different filter modes.
      */
int		canFilterID (unsigned qno, canid_t id);
int		canFilterMask (unsigned qno, canid_t id, canid_t mask);

void		canStart (void);
void		canStop (void);

unsigned	canTxAvail (void);			// return no of empty transmit mailboxes
unsigned	canTransmit (canmsg_t *msg);		// returns mailbox 1-3 or 0
void		canTxComplete (unsigned box);		// transmit completion callback (optional appl defined)
void		canAbort (unsigned box);		// abort mailbox 1-3, or all if 0

unsigned	canRxAvail (unsigned qno);		// return no of full receiver buffers
unsigned	canReceive (canmsg_t *msg, unsigned qno); // returns 0 if no message available

uint32_t	canError (void);			// return error status

#endif
//-----------------------------------------------------------------------------

typedef struct {
    uint32_t idr;
    uint32_t dtr;
    uint32_t dlr;
    uint32_t dhr;
} CAN_MailBox_TypeDef;

//-----------------------------------------------------------------------------

#define CAN_TSR_RQCPn(n)	(CAN_TSR_RQCP0 << ((n)*8))
#define CAN_TSR_TXOKn(n)	(CAN_TSR_TXOK0 << ((n)*8))
#define CAN_TSR_ALSTn(n)	(CAN_TSR_ALST0 << ((n)*8))
#define CAN_TSR_TERRn(n)	(CAN_TSR_TERR0 << ((n)*8))
#define CAN_TSR_ABRQn(n)	(CAN_TSR_ABRQ0 << ((n)*8))
#define CAN_TSR_CODEget(r)	(((r) >> 24) & 3)
#define CAN_TSR_TMEn(n)		(CAN_TSR_TME0 << (n))
#define CAN_TSR_TMEget(r)	(((r) >> 26) & 7)
#define CAN_TSR_LOWn(n)		(CAN_TSR_LOW0 << (n))

#define CAN_RFR_FMPget(r)	((r) & 3)
#define CAN_RFR_FULL		CAN_RF0R_FULL0
#define CAN_RFR_FOVR		CAN_RF0R_FOVR0
#define CAN_RFR_RFOM		CAN_RF0R_RFOM0

#define CAN_ESR_LECget(r)	(((r) >> 4) & 7)
#define CAN_ESR_LECset(n)	((n)<<4)
#define CAN_ESR_LEC_none	0
#define CAN_ESR_LEC_stuff	1
#define CAN_ESR_LEC_form	2
#define CAN_ESR_LEC_acknowledge	3
#define CAN_ESR_LEC_recessive	4
#define CAN_ESR_LEC_dominant	5
#define CAN_ESR_LEC_crc		6
#define CAN_ESR_LEC_software	7

#define CAN_BTR_BRPv(n)		(((n) & 0xFF) << 0)
#define CAN_BTR_TS1v(n)		(((n) & 0x0F) << 16)
#define CAN_BTR_TS2v(n)		(((n) & 0x07) << 20)
#define CAN_BTR_SJWv(n)		(((n) & 0x03) << 24)
#define CAN_BTR(brp,sjw,ts1,ts2) (CAN_BTR_BRPv(brp-1) | CAN_BTR_TS1v(ts1-1) | CAN_BTR_TS2v(ts2-1) | CAN_BTR_SJWv(sjw-1))

#define CAN_IDR_TXRQ		CAN_TI1R_TXRQ
#define CAN_IDR_RTR		CAN_TI1R_RTR
#define CAN_IDR_IDE		CAN_TI1R_IDE
#define CAN_IDR_IDset(id)	((id) << 3)
#define CAN_IDR_IDget(r)	((r) >> 3)

#define CAN_DTR_DCLset(n)     	((n) & 0x0F)
#define CAN_DTR_DCLget(r)     	((r) & 0x0F)
#define CAN_DTR_TGT		CAN_TDT1R_TGT
#define CAN_DTR_FMIget(r)	(((r) >> 8) & 0xFF)     
#define CAN_DTR_TIMEget(r)	((r) >> 16)

//-----------------------------------------------------------------------------
     
#ifdef __cplusplus
}
#endif

#endif //__STM32_CAN_H
